@@ -220,17 +220,17 @@ import Foreign hiding (throwIf, throwIf_, throwIfNull, void)
 import Foreign.C.String
 import Foreign.C.Types (CInt, CShort)
 import System.Posix.Types (Fd(..))
-import System.ZMQ4.Base
 import System.ZMQ4.Internal
-import System.ZMQ4.Error
+import System.ZMQ4.Internal.Base
+import System.ZMQ4.Internal.Error
 
-import qualified Data.ByteString      as SB
-import qualified Data.ByteString.Lazy as LB
-import qualified Data.List.NonEmpty   as S
-import qualified Prelude              as P
-import qualified System.ZMQ4.Base     as B
+import qualified Data.ByteString           as SB
+import qualified Data.ByteString.Lazy      as LB
+import qualified Data.List.NonEmpty        as S
+import qualified Prelude                   as P
+import qualified System.ZMQ4.Internal.Base as B
 
-import GHC.Conc (threadWaitRead, threadWaitWrite)
+import GHC.Conc (threadWaitRead)
 import GHC.Generics(Generic)
 
 -----------------------------------------------------------------------------
@@ -792,14 +792,15 @@ disconnect sock str = onSocket "disconnect" sock $
 -- by default. Still 'send' is blocking the thread as long as the message
 -- can not be queued on the socket using GHC's 'threadWaitWrite'.
 send :: Sender a => Socket a -> [Flag] -> SB.ByteString -> IO ()
-send sock fls val = bracket (messageOf val) messageClose $ \m ->
-  onSocket "send" sock $ \s ->
-    retry "send" (waitWrite sock) $
+send sock fls val = bracketOnError (messageOf val) messageClose $ \m -> do
+    onSocket "send" sock $ \s ->
+        retry "send" (waitWrite sock) $
 #ifdef mingw32_HOST_OS
-          c_zmq_sendmsg s (msgPtr m) (combineFlags fls)
+            c_zmq_sendmsg s (msgPtr m) (combineFlags fls)
 #else
-          c_zmq_sendmsg s (msgPtr m) (combineFlags (DontWait : fls))
+            c_zmq_sendmsg s (msgPtr m) (combineFlags (DontWait : fls))
 #endif
+    messageFree m
 
 -- | Send the given 'LB.ByteString' over the socket
 -- (cf. <http://api.zeromq.org/4-0:zmq-sendmsg zmq_sendmsg>).
@@ -812,14 +813,15 @@ send sock fls val = bracket (messageOf val) messageClose $ \m ->
 -- by default. Still 'send'' is blocking the thread as long as the message
 -- can not be queued on the socket using GHC's 'threadWaitWrite'.
 send' :: Sender a => Socket a -> [Flag] -> LB.ByteString -> IO ()
-send' sock fls val = bracket (messageOfLazy val) messageClose $ \m ->
-  onSocket "send'" sock $ \s ->
-    retry "send'" (waitWrite sock) $
+send' sock fls val = bracketOnError (messageOfLazy val) messageClose $ \m -> do
+    onSocket "send'" sock $ \s ->
+        retry "send'" (waitWrite sock) $
 #ifdef mingw32_HOST_OS
-          c_zmq_sendmsg s (msgPtr m) (combineFlags fls)
+            c_zmq_sendmsg s (msgPtr m) (combineFlags fls)
 #else
-          c_zmq_sendmsg s (msgPtr m) (combineFlags (DontWait : fls))
+            c_zmq_sendmsg s (msgPtr m) (combineFlags (DontWait : fls))
 #endif
+    messageFree m
 
 -- | Send a multi-part message.
 -- This function applies the 'SendMore' 'Flag' between all message parts.
@@ -896,8 +898,7 @@ monitor es ctx sock = do
 #else
             c_zmq_recvmsg s (msgPtr m) (flagVal dontWait)
 #endif
-        ptr <- c_zmq_msg_data (msgPtr m)
-        evt <- peek ptr
+        evt <- peekZMQEvent (msgPtr m)
         str <- receive soc
         return . Just $ eventMessage str evt
 
@@ -948,38 +949,38 @@ toEvents e = foldl' (\es f -> f e es) [] tests
 retry :: String -> IO () -> IO CInt -> IO ()
 retry msg wait act = throwIfMinus1RetryMayBlock_ msg act wait
 
-wait' :: (Fd -> IO ()) -> ZMQPollEvent -> Socket a -> IO ()
+wait' :: ZMQPollEvent -> Socket a -> IO ()
 #ifdef mingw32_HOST_OS
-wait' _ _ _ = return ()
+wait' _ _ = return ()
 #else
-wait' w f s = do
-    fd <- getIntOpt s B.filedesc 0
-    w (Fd fd)
-    evs <- getInt32Option B.events s
-    unless (testev evs) $
-        wait' w f s
+wait' p s = do
+    e <- getInt32Option B.events s
+    unless (testev e) $ do
+        fd <- getIntOpt s B.filedesc 0
+        threadWaitRead (Fd fd)
+        wait' p s
   where
-    testev e = e .&. fromIntegral (pollVal f) /= 0
+    testev e = e .&. fromIntegral (pollVal p) /= 0
 #endif
 
 -- | Wait until data is available for reading from the given Socket.
 -- After this function returns, a call to 'receive' will essentially be
 -- non-blocking.
 waitRead :: Socket a -> IO ()
-waitRead = wait' threadWaitRead pollIn
+waitRead = wait' pollIn
 
 -- | Wait until data can be written to the given Socket.
 -- After this function returns, a call to 'send' will essentially be
 -- non-blocking.
 waitWrite :: Socket a -> IO ()
-waitWrite = wait' threadWaitWrite pollOut
+waitWrite = wait' pollOut
 
 -- | Starts built-in 0MQ proxy
 -- (cf. <http://api.zeromq.org/4-0:zmq-proxy zmq_proxy>)
 --
 -- Proxy connects front to back socket
 --
--- Before calling proxy all sockets should be binded
+-- Before calling proxy all sockets should be bound
 --
 -- If the capture socket is not Nothing, the proxy  shall send all
 -- messages, received on both frontend and backend, to the capture socket.
